/*****************************************************************************************
 ** File     : PasswordFilter.cpp
 ** Date     : 03/09/2002
 ** Author   : Osama Dengler <osama@denglernet.de>
 ** Abstract : Windows NT / Unix password synchronization filter package
 **
 ** Copyright (C) Osama Dengler 2002
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation; either version 2 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program; if not, write to the Free Software
 ** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *****************************************************************************************/

// !!! _UNICODE and UNICODE must be defined !!!

#include "stdafx.h"
#include "PasswordFilter.h"
#include "Windows.h"
#include "Debug.h"
#include "NtSecPkg.h"

HINSTANCE hThisDll;
extern FILE* logfile;
PTSTR pLdapServer=NULL;
PTSTR pBindDN=NULL;
PTSTR pBindPW=NULL;
PTSTR pBaseDN=NULL;
BOOL  bUseSSL=FALSE;
BOOL  bForceLDAPv2=FALSE;
BOOL  bForceLDAPv3=FALSE;
int   iLdapVersion=LDAP_VERSION3;

/*************************************************************************************************
 ** Function  : DllMain
 ** Purpose   : Library initialisation entry point
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOL APIENTRY DllMain( HANDLE hModule, DWORD dwReason, LPVOID lpReserved )
{
#ifdef _DEBUG
    SYSTEMTIME st;
	TCHAR TimeString[30];
	TCHAR Filename[MAX_PATH];
	TCHAR ProcessName[MAX_PATH];

	GetSystemTime(&st);
	_stprintf(Filename,L"C:\\PasswordFilter.%s.log",DumpSysDateTimeShort(&st,(TCHAR*)&TimeString));

	GetModuleFileName(NULL,ProcessName,MAX_PATH);
#endif
    switch (dwReason)
    {
        case DLL_PROCESS_ATTACH: hThisDll = (HINSTANCE)hModule;
								 DisableThreadLibraryCalls (hThisDll);
#ifdef _DEBUG
								 logfile=_tfopen(Filename,TEXT("a+"));
								 if(logfile==NULL) return(FALSE);

								 DEBUGMSG("Log begin\n");
								 DEBUGMSG("Date / Time : %S\n",DumpSysDateTime(&st,(TCHAR*)&TimeString));
								 DEBUGMSG("PasswordFilter.dll - Timestamp : %s/%s\n",__DATE__,__TIME__);
								 DEBUGMSG("Calling process: %S\n",ProcessName);
#endif
								 break;
#ifdef _DEBUG
        case DLL_PROCESS_DETACH: DEBUGMSG("----------------------------------------------------------------------------\n");
								 DEBUGMSG("%S - Log end\n",DumpSysDateTime(&st,(TCHAR*)&TimeString));
								 DEBUGMSG("Process %s detached from PasswordFilter.dll\n",ProcessName);
								 DEBUGMSG("Library instance unloaded.\n");
								 DEBUGMSG("----------------------------------------------------------------------------\n");
								 fclose(logfile);
								 logfile = NULL;
								 break;
#endif
    }

    return(TRUE);
}

/*************************************************************************************************
 ** Function  : InitializeChangeNotify
 ** Purpose   : Initialize the package
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOLEAN __stdcall InitializeChangeNotify()
{
	DEBUGMSG("----------------------------------------------------------------------------\n");
	DEBUGMSG("+InitializeChangeNotify\n");

	/*
	 * Read the settings from the registry. If the registry is not accessible, fail.
	 */
	if(!ReadRegistrySettings())
	{
		 DEBUGMSG("   Unable to read registry settings\n");
		 return(FALSE);
	}

	/*
	 * Set the desired LDAP protocol version. If the user forces a certain version,
	 * stick to this, otherwise ask the server if it supports LDAPv3.
	 */
	if(bForceLDAPv2)
	{
		iLdapVersion=LDAP_VERSION2;
		DEBUGMSG("   Forced LDAPv2\n");
	}
	else if(bForceLDAPv3)
	{
		iLdapVersion=LDAP_VERSION3;
		DEBUGMSG("   Forced LDAPv3\n");
	}
	else iLdapVersion=LdapGetProtocolVersion();

	if((LDAP_VERSION2 != iLdapVersion) && (LDAP_VERSION3 != iLdapVersion))
	{
		DEBUGMSG("      Error: unsupported LDAP-Version (LDAPv%u)\n", iLdapVersion);
		DEBUGMSG("-InitializeChangeNotify\n");
		return(FALSE);
	}

	DEBUGMSG("   LDAP Version is %u\n", iLdapVersion);
	DEBUGMSG("-InitializeChangeNotify\n");
	DEBUGMSG("----------------------------------------------------------------------------\n");

	return(TRUE);
}

/*************************************************************************************************
 ** Function  : PasswordChangeNotify
 ** Purpose   : Change the password
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
NTSTATUS __stdcall PasswordChangeNotify( PUNICODE_STRING UserName,
										 ULONG RelativeId,
										 PUNICODE_STRING NewPassword)
{
	DEBUGMSG("----------------------------------------------------------------------------\n");
	DEBUGMSG("+PasswordChangeNotify\n");
	DEBUGMSG("UID : %S\n", UserName->Buffer);
	DEBUGMSG("RID : %u\n", RelativeId);
	DEBUGMSG("PWD : %S\n", NewPassword->Buffer);

	LdapChangePassword(UserName->Buffer, NewPassword->Buffer);

	DEBUGMSG("-PasswordChangeNotify\n");
	DEBUGMSG("----------------------------------------------------------------------------\n");
	return(STATUS_SUCCESS);
}

/*************************************************************************************************
 ** Function  : PasswordFilter
 ** Purpose   : Change the password
 ** Author    : O. Dengler
 ** History   : 03/09/2002 : 1st
 ************************************************************************************************/
BOOLEAN __stdcall PasswordFilter( PUNICODE_STRING AccountName,
								  PUNICODE_STRING FullName,
								  PUNICODE_STRING Password,
								  BOOLEAN SetOperation)
{
	PTSTR pAccountName=ZeroTerminateString(AccountName);
	PTSTR pFullName=ZeroTerminateString(FullName);
	PTSTR pPassword=ZeroTerminateString(Password);

	DEBUGMSG("----------------------------------------------------------------------------\n");
	DEBUGMSG("+PasswordFilter\n");
	DEBUGMSG("UID  : %S\n", pAccountName);
	DEBUGMSG("Full : %S\n", pFullName);
	DEBUGMSG("PWD  : %S\n", pPassword);
	DEBUGMSG("OP   : %s\n", SetOperation ? "SetPassword" : "ChangePassword");
	DEBUGMSG("-PasswordFilter\n");
	DEBUGMSG("----------------------------------------------------------------------------\n");

	if(NULL != pAccountName) LocalFree(pAccountName);
	if(NULL != pFullName) LocalFree(pFullName);
	if(NULL != pPassword) LocalFree(pPassword);

	return(TRUE);
}

/*************************************************************************************************
 ** Function  : ZeroTerminateString
 ** Purpose   : UNICODE_STRINGs are not granted to be zero-terminated. Generate a local copy and
 **             terminate it with zero.
 ** Author    : O. Dengler
 ** History   : 03/13/2002 : 1st
 ************************************************************************************************/
PWSTR ZeroTerminateString(PUNICODE_STRING pUnicodeString)
{
	PWSTR pDst=NULL;
	
	DEBUGMSG("   +ZeroTerminateString\n");

	if(NULL == pUnicodeString)
	{
		DEBUGMSG("   Warning: NULL unicode string. Returning NULL pointer.\n");
		return(NULL);
	}

	DEBUGMSG("      UnicodeString (%u,%u) = \"%S\"\n", pUnicodeString->Length,
													   pUnicodeString->MaximumLength,
													   pUnicodeString->Buffer);
	if(pUnicodeString->Length > 0)
	{
		pDst=(PWSTR)LocalAlloc(LPTR, (pUnicodeString->Length +1 ) * sizeof(WCHAR));

		wcsncpy(pDst, pUnicodeString->Buffer, pUnicodeString->Length+1);
	}

	DEBUGMSG("      Zero Terminated String = \"%S\"\n", pDst==NULL ? _T("NULL") : pDst);
	DEBUGMSG("   -ZeroTerminateString\n");
	return(pDst);
}

/*************************************************************************************************
 ** Function  : GetRegValString
 ** Purpose   : Get a REG_SZ registry key value
 ** Author    : O. Dengler
 ** History   : 03/20/2002 : 1st
 ************************************************************************************************/
PTSTR GetRegValString(HKEY hRegKey, PTSTR pValueName)
{
	DWORD dwKeyType, dwBufSize;
	PTSTR pData=NULL;

	/*
	 * Determine the amount of memory necessary to hold the string
	 */
	if(RegQueryValueEx(	hRegKey,
						pValueName,
						NULL,
						&dwKeyType,
						NULL,
						&dwBufSize) != ERROR_SUCCESS)
		/*
		 * return FAILURE.
		 */
		return(NULL);

	/*
	 * Allocate the destination buffer
	 */
	pData=(PTSTR)LocalAlloc(LPTR, dwBufSize);
	if(NULL == pData)
		/*
		 * return FAILURE.
		 */
		return(NULL);

	/*
	 * Query the string into the destination buffer
	 */
	if(RegQueryValueEx(	hRegKey,
						pValueName,
						NULL,
						&dwKeyType,
						(PBYTE)pData,
						&dwBufSize) != ERROR_SUCCESS)
	{
		/*
		 * Free previously allocated memory and return FAILURE.
		 */
		LocalFree(pData);
		return(NULL);
	}

	/*
	 * return the string.
	 */
	return(pData);
}

/*************************************************************************************************
 ** Function  : GetRegValBool
 ** Purpose   : Get a BOOLean registry key value
 ** Author    : O. Dengler
 ** History   : 03/20/2002 : 1st
 ************************************************************************************************/
BOOL GetRegValBool(HKEY hRegKey, PTSTR pValueName, PBOOL pData)
{
	DWORD dwKeyType, dwBufSize;

	/*
	 * Query the registry key value
	 */
	if(RegQueryValueEx(	hRegKey,
						pValueName,
						NULL,
						&dwKeyType,
						(PBYTE)pData,
						&dwBufSize) != ERROR_SUCCESS)
		/*
		 * return FAILURE.
		 */
		return(FALSE);

	/*
	 * return SUCCESS.
	 */
	return(TRUE);
}

/*************************************************************************************************
 ** Function  : ReadRegistrySettings
 ** Purpose   : Read the OpenAuth PasswordFilter registry settings.
 ** Author    : O. Dengler
 ** History   : 03/20/2002 : 1st
 ************************************************************************************************/
BOOL ReadRegistrySettings(void)
{
	HKEY hRegKey;
	
	DEBUGMSG("+ReadRegistrySettings\n");

	/*
	 * Open the registry key
	 */
	if(RegOpenKeyEx(HKEY_LOCAL_MACHINE, REG_KEY, 0,  KEY_READ, &hRegKey) != ERROR_SUCCESS)
	{
		DEBUGMSG("   ERROR : unable to open registry key.\n");
		DEBUGMSG("-ReadRegistrySettings\n");
		return(FALSE);
	}
	
	/*
	 * Get the name of the LDAP server.
	 */
	pLdapServer=GetRegValString(hRegKey, _T("LDAP Server"));
	if(NULL == pLdapServer)
	{
		DEBUGMSG("   Unable to read LDAP Server from registry.\n");
		goto ErrorReturn;
	}
	DEBUGMSG("   LDAP Server = %S\n", pLdapServer);

	/*
	 * Get the search base DN.
	 */
	pBaseDN=GetRegValString(hRegKey, _T("Base DN"));
	if(NULL == pBaseDN)
	{
		DEBUGMSG("   Unable to read Base DN from registry.\n");
		goto ErrorReturn;
	}
	DEBUGMSG("   Base DN = %S\n", pBaseDN);
 
	/*
	 * Get the bind DN
	 */
	pBindDN=GetRegValString(hRegKey, _T("Bind DN"));
	if(NULL == pBindDN)
	{
		DEBUGMSG("   Unable to read Bind DN from registry.\n");
		goto ErrorReturn;
	}
	DEBUGMSG("   Bind DN = %S\n", pBindDN);

	/*
	 * Get the bind PW.
	 */
	pBindPW=GetRegValString(hRegKey, _T("Bind PW"));
	if(NULL == pBindPW)
	{
		DEBUGMSG("   Unable to read Bind PW from registry.\n");
		goto ErrorReturn;
	}
	DEBUGMSG("   Bind PW = %S\n", pBindPW);

	/*
	 * Get the SSL flag.
	 */
	if(!GetRegValBool(hRegKey, _T("Use SSL"), &bUseSSL))
	{
		DEBUGMSG("Unable to read SSL flag from registry.\n");
		goto ErrorReturn;
	}
	DEBUGMSG("   SSL flag = %s\n", bUseSSL ? "TRUE" : "FALSE");

	/*
	 * Get the LDAPv2 flag.
	 */
	if(!GetRegValBool(hRegKey, _T("Force LDAPv2"), &bForceLDAPv2))
	{
		DEBUGMSG("Unable to read LDAPv2 flag from registry.\n");
		goto ErrorReturn;
	}
	DEBUGMSG("   LDAPv2 flag = %s\n", bForceLDAPv2 ? "TRUE" : "FALSE");

	/*
	 * Get the LDAPv3 flag.
	 */
	if(!GetRegValBool(hRegKey, _T("Force LDAPv3"), &bForceLDAPv3))
	{
		DEBUGMSG("Unable to read LDAPv3 flag from registry.\n");
		goto ErrorReturn;
	}
	DEBUGMSG("   LDAPv3 flag = %s\n", bForceLDAPv3 ? "TRUE" : "FALSE");
	
	/*
	 * Close the registry key.
	 */
	RegCloseKey(hRegKey);

	/*
	 * Return SUCCESS.
	 */
	DEBUGMSG("-ReadRegistrySettings\n");
	return(TRUE);

	/*
	 * In case one or all registry keys could not be queried, release all memory
	 * and fail. We zero all settings out so that they won't appear in eventual 
	 * coredumps.
	 */
ErrorReturn:
	if(NULL != pLdapServer)
	{
		ZeroMemory(pLdapServer, _tcslen(pLdapServer) * sizeof(TCHAR));
		LocalFree(pLdapServer);
	}

	if(NULL != pBaseDN)
	{
		ZeroMemory(pBaseDN, _tcslen(pBaseDN) * sizeof(TCHAR));
		LocalFree(pBaseDN);
	}

	if(NULL != pBindDN)
	{
		ZeroMemory(pBindDN, _tcslen(pBindDN) * sizeof(TCHAR));
		LocalFree(pBindDN);
	}

	if(NULL != pBindPW)
	{
		ZeroMemory(pBindPW, _tcslen(pBindPW) * sizeof(TCHAR));
		LocalFree(pBindPW);
	}

	/*
	 * Close the key.
	 */
	RegCloseKey(hRegKey);

	/*
	 * Return FAILURE.
	 */
	DEBUGMSG("-ReadRegistrySettings\n");
	return(FALSE);
}